include machine.e
--include safe.e as safe


global constant
	PPUCTR1		= #2000,
	PPUCTR2		= #2001,
	PPUSTAT		= #2002,
	SPRRADR		= #2003,
	SPRRIO		= #2004,
	VRAMADR1	= #2005,
	VRAMADR2	= #2006,
	VRAMIO		= #2007,
	SPRDMA		= #4014

global constant
	HORIZONTAL	= 0,
	VERTICAL	= 1
	
	
-- REGISTERS
global integer
	reg_A,		-- accumulator
	reg_X,		-- index register
	reg_Y,		-- secondary index register
	reg_P,		-- processor status register
	reg_S,		-- stack register
	reg_PC		-- program counter


global integer
	hScroll,	-- horizontal scroll value
	vScroll,	-- vertical scroll value
	scrollWrite,	-- scroll register write counter
	firstVRAMRead,
	vramPtr,
	buttonInx,
	logfile,
	vramMirroring,
	mapper,
 	nPrgPages,		-- number of 16k PRG-ROM pages
        nChrPages,		-- number of 8k CHR-ROM pages
 	lines,
	scanline,
	chrRomPage,
	prgRomPage,
	in_vblank,
	refreshPatternTable,
	scrollLP,
	
	mmc1Buffer,
	mmc1BufferMask,
	mmc1OneScreenMirroring,
	mmc1PrgSwitchingArea,
	mmc1PrgSwitchingSize,
	mmc1Reg0Bit4,
	mmc1VROMSelect0,
	mmc1VROMSelect1,
	mmc1Select256K0,
	mmc1Select256K1,
	mmc1SelectROMBank
	
	

global atom
    	RAM,
     	RAMM1,
     	RAMM2,
     	RAMM3,
     	EXROM,
     	SRAM,
     	PRGROM1,
     	PRGROM2,
     	VRAM,
     	SPRRAM,
     	PATTBL0,
     	PATTBL1,
     	chrMem,
     	prgMem

global atom cycle

global sequence buttons,
		scrollList

integer sprRAMOffs
sequence buttonCopy

	
	
global procedure init_memory(atom memory)
	RAM   = memory
	RAMM1 = memory + #0800
	RAMM2 = memory + #1000
	RAMM3 = memory + #1800
	EXROM = memory + #4020
	SRAM  = memory + #6000
	--PRGROM1 = memory + #8000
	--PRGROM2 = memory + #C000
	VRAM = memory + #10000
	SPRRAM = memory + #14000

	mem_set(RAM,0,#14100)
	sprRAMOffs = 0
	hScroll = 0
	vScroll = 0
	scrollWrite = 0
	firstVRAMRead = 1
	vramPtr = 0
	buttonInx = 0
	buttons = repeat(0,8)
	buttonCopy = buttons
	--mapper = 0
	scanline = 0

	mmc1Buffer	= 0
	mmc1BufferMask	= 1
	mmc1OneScreenMirroring = 0
	mmc1PrgSwitchingArea = 0
	mmc1PrgSwitchingSize = 0
	mmc1Reg0Bit4 = 0
	mmc1VROMSelect0 = 0
	mmc1VROMSelect1 = 0
	mmc1Select256K0 = 0
	mmc1Select256K1 = 0
	mmc1SelectROMBank = 0
end procedure



-- Soft reset
global procedure reset_memory()
	--sprRAMOffs = 0
	--hScroll = 0
	--vScroll = 0
	--scrollWrite = 0
	--firstVRAMRead = 1
	vramPtr = 0
	buttonInx = 0
	buttons = repeat(0,8)
	buttonCopy = buttons
	--mapper = 0
	scanline = 0

	mmc1Buffer	= 0
	mmc1BufferMask	= 1
	mmc1OneScreenMirroring = 0
	mmc1PrgSwitchingArea = 0
	mmc1PrgSwitchingSize = 0
	mmc1Reg0Bit4 = 0
	mmc1VROMSelect0 = 0
	mmc1VROMSelect1 = 0
	mmc1Select256K0 = 0
	mmc1Select256K1 = 0
	mmc1SelectROMBank = 0
end procedure



-- Read one byte (8 bits) from PC
global function fetch_byte()
	if reg_PC<#C000 then
		reg_PC += 1
		return peek(reg_PC+PRGROM1-#8001)
	end if
	reg_PC += 1
	return peek(reg_PC+PRGROM2-#C001)
end function



-- Read one word (16 bits) from PC
global function fetch_word()
	integer a
	if reg_PC<#C000 then
		a = PRGROM1+reg_PC-#8000
	else
		a = PRGROM2+reg_PC-#C000
	end if	
	reg_PC += 2
	return peek(a)+peek(a+1)*#100
end function



global function read_byte(integer addr)
	integer b
	
	--addr = and_bits(addr,#FFFF)
	
	--if addr<#2000 then
	--	addr = and_bits(addr,#7FF)
	--elsif addr>=#2008 and addr<#4000 then
	--	addr = and_bits(addr,7)+#2000
	--end if
	
	if addr<#2000 then
		b = peek(RAM+and_bits(addr,#7FF))

	elsif addr<#4000 then
		addr = and_bits(addr,7)+#2000
	
		if addr=PPUSTAT then
			b = peek(addr+RAM)
			poke(addr+RAM,and_bits(b,#3F))
			poke(RAM+#2005,0)
			poke(RAM+#2006,0)
			scrollWrite = 0
			vramPtr = 0

			if not in_vblank then
				if not and_bits(peek(RAM+#2000),#80) then
					poke(RAM+#2000,and_bits(peek(RAM+#2000),#FC))
				end if
			end if
		
		elsif addr=SPRRIO then
			b = peek(SPRRAM+peek(RAM+SPRRADR))
	
		elsif addr=VRAMIO then
   			b = peek(VRAM+vramPtr)
   			if not firstVRAMRead then
   				vramPtr += 1 + (and_bits(peek(RAM+PPUCTR1),4)!=0)*31
   			else
   				vramPtr += 1
   			end if
   			vramPtr = and_bits(vramPtr,#3FFF)
   			firstVRAMRead = 0
   		end if
   		
	elsif addr=#4016 then
		if buttonInx>0 and buttonInx<9 then
			b = #40+buttonCopy[buttonInx]
		elsif buttonInx=17 or buttonInx=20 then
			b = 1	-- joypad 1 connected
		else
			b = 0
		end if
		if buttonInx>0 then
			buttonInx += 1
		end if
		if buttonInx>24 then
			buttonInx = 0
		end if

	elsif addr=#4017 then
		b = 0
		
	elsif addr>=#7000 and addr<#7200 then
		b = peek(RAM+addr)

	elsif addr>=#8000 then
		if addr<#C000 then
			b = peek(PRGROM1+addr-#8000)
		else
			b = peek(PRGROM2+addr-#C000)
		end if

	else 
		printf(logfile,"Illegal read from %04x at PC = %04x\n",{addr,reg_PC})
		b = 0
	end if

	return b
end function



global function read_word(integer addr)
	integer i

	i = read_byte(addr)
	-- Don't increase high byte of address on page wrap
	addr = and_bits(addr+1,#FF)+and_bits(addr,#FF00)
	--addr += 1
	
	return i + read_byte(addr)*#100
end function


global function read_word2(integer addr)
	integer i,j
	
	if addr<#8000 then
		i = peek(RAM+addr)
		return i + peek(RAM+addr+1)*#100
	elsif addr<#C000 then
		i = peek(PRGROM1+addr-#8000)
		return i + peek(PRGROM1+addr-#7FFF)*#100
	else
		i = peek(PRGROM2+addr-#C000)
		return i + peek(PRGROM2+addr-#BFFF)*#100
	end if		
end function



global procedure write_byte(integer addr, integer b)
	--addr = and_bits(addr,#FFFF)
	
	--if addr<#2000 then
	--	addr = and_bits(addr,#7FF)
	--elsif addr>=#2008 and addr<#4000 then
	--	addr = and_bits(addr,7)+#2000
	--end if

	if addr<#2000 then
		poke(RAM+and_bits(addr,#7FF),b)

	elsif addr<#4000 then
		addr = and_bits(addr,7)+#2000
		
		if addr=PPUCTR1 or addr=PPUCTR2 then
			poke(RAM+addr,b)
	
		elsif addr=SPRRADR then
			poke(RAM+SPRRADR,b)
		
		elsif addr=SPRRIO then
			sprRAMOffs = peek(RAM+SPRRADR)
			poke(SPRRAM+sprRAMOffs,b)
			poke(RAM+SPRRADR,sprRAMOffs+1)
		
		elsif addr=VRAMADR1 then
			if scrollWrite=1 then
				vScroll = b
				if vScroll>239 then
					vScroll = 0 ---= 240
				end if
				scrollWrite = 0
			else
				hScroll = b
				scrollWrite = 1
			end if
	
		elsif addr=VRAMADR2 then
        		vramPtr = and_bits(vramPtr,#FF)*#100 + b
        		poke(RAM+VRAMIO,0)
        	 	firstVRAMRead = 1
        
        	elsif addr=VRAMIO then
			if not and_bits(peek(RAM+PPUSTAT),#10) then
		         	vramPtr = and_bits(vramPtr,#3FFF)
	
				if vramPtr>=#2000 then
		         		poke(VRAM+vramPtr,b)
		         	elsif mapper=2 then
		         		poke(PATTBL0+vramPtr,b)
		         		refreshPatternTable = 1
		         	end if
	         	
		         	if vramPtr>=#2000 and vramPtr<#3000 then
		         		if vramMirroring=VERTICAL then
		         			if and_bits(vramPtr,#800) then
		         				poke(VRAM+vramPtr-#800,b)
		         			else
		         				poke(VRAM+vramPtr+#800,b)
		         			end if
		         		elsif vramMirroring=HORIZONTAL then
		         			if and_bits(vramPtr,#400) then
		         				poke(VRAM+vramPtr-#400,b)
		         			else
		         				poke(VRAM+vramPtr+#400,b)
		         			end if
		         		end if
		         	end if
	         	
		         	-- Palette mirroring
        	     		if vramPtr>=#3F00 and vramPtr<=#3FFF then
        		     		if not and_bits(vramPtr,3) then
						addr = and_bits(vramPtr,#3F0F)
						poke(VRAM+addr,b)
						poke(VRAM+addr+#10,b)
					end if
				end if
			
		         	vramPtr += 1 + (and_bits(peek(RAM+PPUCTR1),4)!=0)*31
		         	vramPtr = and_bits(vramPtr,#3FFF)
			end if
        	end if
        	
	elsif addr=SPRDMA then
		-- Transfer 256 bytes from b*#100 to spr-ram
		mem_copy(SPRRAM,RAM+(and_bits(b,7)*#100),#100)
		cycle += 513

	elsif addr=#4016 then
		if not and_bits(b,1) then
			if and_bits(peek(RAM+#4016),1) then
				--puts(1,"Reading joypad 1\n")
				-- read joypad 1
				buttonCopy = buttons
				buttonInx = 1
			end if
		end if
		poke(RAM+#4016,b)
	
	elsif addr=#4017 then

		
	elsif addr>=#8000 then
		if mapper=1 then
			if mmc1BufferMask=#80 then
				if b then
					if addr<#A000 then
						mmc1OneScreenMirroring = 0
						mmc1PrgSwitchingArea = 0
						mmc1PrgSwitchingSize = 0
						mmc1Reg0Bit4 = 0
					elsif addr<#C000 then
						mmc1VROMSelect0 = 0
						mmc1Select256K0 = 0
					elsif addr<#E000 then
						mmc1VROMSelect1 = 0
						mmc1Select256K1 = 0
					else
						mmc1SelectROMBank = 0
					end if
				end if
			elsif mmc1BufferMask<=#10 then
				if b then
					mmc1Buffer = or_bits(mmc1Buffer,mmc1BufferMask)
				end if
				if mmc1BufferMask=#10 then
					if addr<#A000 then
						vramMirroring = and_bits(mmc1Buffer,1)
						mmc1OneScreenMirroring = and_bits(mmc1Buffer,2)
						mmc1PrgSwitchingArea = and_bits(mmc1Buffer,4)
						mmc1PrgSwitchingSize = and_bits(mmc1Buffer,8)
						mmc1Reg0Bit4 = and_bits(mmc1Buffer,16)
					elsif addr<#C000 then
						mmc1VROMSelect0 = and_bits(mmc1Buffer,15)
						mmc1Select256K0 = and_bits(mmc1Buffer,16)

						if nChrPages then
						if mmc1Reg0Bit4 then
							PATTBL0 = chrMem+mmc1VROMSelect0*#1000
							refreshPatternTable = 1
						else
							PATTBL0 = chrMem+mmc1VROMSelect0*#2000
							PATTBL1 = chrMem+mmc1VROMSelect0*#2000 + #1000
							refreshPatternTable = 1
						end if
						end if						
					elsif addr<#E000 then
						mmc1VROMSelect1 = and_bits(mmc1Buffer,15)
						mmc1Select256K1 = and_bits(mmc1Buffer,16)
						
						if nChrPages then
						if mmc1Reg0Bit4 then
							PATTBL1 = chrMem+mmc1VROMSelect1*#2000 + #1000
							refreshPatternTable = 1
						end if
						end if
					else
						mmc1SelectROMBank = and_bits(mmc1Buffer,nPrgPages-1)
						if mmc1PrgSwitchingSize then
							-- Switch 16k
							if mmc1PrgSwitchingArea then
								-- ..at $8000
								printf(1,"Switching bank %d at $8000\n",mmc1SelectROMBank)
								PRGROM1 = prgMem + mmc1SelectROMBank*#4000
							else
								-- ..at $C000
								printf(1,"Switching bank %d at $C000\n",mmc1SelectROMBank)
								PRGROM2 = prgMem + mmc1SelectROMBank*#4000
							end if
						else
							-- Switch 32k at $8000
							PRGROM1 = prgMem + floor(mmc1SelectROMBank/2)*#4000
							PRGROM1 = prgMem + floor(mmc1SelectROMBank/2)*#4000 + #4000
						end if
					end if
				end if
			end if
			mmc1BufferMask *= 2
			if mmc1BufferMask>#FF then
				mmc1Buffer = 0
				mmc1BufferMask = 1
			end if
			
		
		elsif mapper=2 then
			PRGROM1 = prgMem + b*#4000
			
			
		elsif mapper=3 then
			if and_bits(addr,#7FFF) and b!=chrRomPage then
				chrRomPage = b
				PATTBL0 = chrMem + and_bits(b,15)*#2000
				PATTBL1 = PATTBL0 + #1000
				refreshPatternTable = 1
			end if
		
		elsif mapper=66 then
			PRGROM1 = prgMem + and_bits(floor(b/16),3)*#4000
			PRGROM2 = PRGROM1 + #4000
			printf(1,"Switching in bank %d at $8000 ($%04x)\n",{and_bits(floor(b/16),3),addr})
			if b!=chrRomPage then
				chrRomPage = and_bits(b,3)
				PATTBL0 = chrMem + chrRomPage*#2000
				PATTBL1 = PATTBL0 + #1000
				refreshPatternTable = 1
			end if
			
		else
			puts(logfile,"Attempting to write to ROM!\n")
		end if
	else
		--poke(RAM+addr,b)	-- Temp. hack
	end if
end procedure



-- Return an 8-bit relative address in 2's complement
global function rel8()
	integer addr
	
	addr = fetch_byte()
	return addr - ((addr>=#80)*256)
end function


-- Zero-paged X-indexed
global function zpx8()
	return and_bits(fetch_byte()+reg_X,#FF)
end function


global function zpy8()
	return and_bits(fetch_byte()+reg_Y,#FF)
end function


global function zpx16()
	return and_bits(fetch_word()+reg_X,#FFFF)
end function


global function zpy16()
	return and_bits(fetch_word()+reg_Y,#FFFF)
end function


-- Immediate
global function imm8()
	return fetch_byte()
end function


-- Indirect
global function in16()
	return read_word(fetch_word())
end function


-- Indirect, prefix X-indexed
global function inx8pre()
	return read_word(and_bits(fetch_byte()+reg_X,#FF))
end function


-- Indirect, postfix Y-indexed
global function iny8post()
	return read_word(fetch_byte()) + reg_Y
end function


global function abs8(integer addr)
	return and_bits(addr,#FF)
end function


global function abs16(integer addr)
	return and_bits(addr,#FFFF)
end function


-- Push b onto the stack (located on page 1 in CPU memory space)
global procedure push(integer b)
	poke(reg_S+RAM+#100,b)
	reg_S = and_bits(reg_S-1,#FF)
end procedure


-- Push w onto the stack (located on page 1 in CPU memory space)
global procedure push_word(integer w)
	poke(reg_S+RAM+#100,floor(w/#100))
	reg_S = and_bits(reg_S-1,#FF)
	poke(reg_S+RAM+#100,w)
	reg_S = and_bits(reg_S-1,#FF)
end procedure


-- Pull a byte from the stack
global function pull()
	reg_S = and_bits(reg_S+1,#FF)
	return peek(reg_S+RAM+#100)
end function
